@@ -2,6 +2,8 @@
 
 When importing objects from modules, list each alphabetically.
 
+Modules in the top-level package can import from utils but not from cam or parts_linkers packages.
+
 """
 
 from .main import BasicAssembly, BasicLinker, BasicPart, BasicUTRRBSLinker, seqrec2part
@@ -23,6 +25,7 @@ from .cam import (
     BuildEncoder,
     build_digest,
     new_part_resuspension,
-    export_echo_assembly,
+    export_echo_assembly_instructions,
+    export_echo_clips_instructions,
     pdf_instructions,
 )
