@@ -1,15 +1,20 @@
-"""Main module for basicsynbio."""
-from basicsynbio.utils import _easy_seqrec
+"""Main module for basicsynbio.
+
+This module can import from utils but not from cam or parts_linkers packages.
+
+"""
+from basicsynbio.utils import _easy_seqrec, p3_seqrec, OLIGO_ANNOTATIONS
 from basicsynbio.decorators import addargs2docs, ArgDescription
 from Bio import SeqUtils, SeqIO
 from Bio.Restriction.Restriction import BsaI
 from Bio.Seq import Seq
 from Bio.SeqRecord import SeqRecord
 from Bio.SeqFeature import SeqFeature, FeatureLocation
-from Bio.SeqUtils.CheckSum import seguid
 from collections import Counter
+from dataclasses import dataclass
 import datetime
 import hashlib
+from primer3.bindings import designPrimers
 from typing import Union, Tuple
 import warnings
 
@@ -47,11 +52,18 @@ class CommonArgDocs:
     )
 
 
+@dataclass
+class DomesticatingPrimers:
+    left_primer: SeqRecord
+    right_primer: SeqRecord
+    data: dict
+
+
 class BasicPart(SeqRecord):
     """Class for BASIC DNA assembly parts.
 
     A DNA sequence joined with other BasicParts via :py:class:`BasicLinker`
-    instances when initialising :py:class:`BasicAssembly` objects. All
+    instances when initializing :py:class:`BasicAssembly` objects. All
     sequences must contain intergated prefix and suffix sequences.
 
     Attributes:
@@ -90,26 +102,15 @@ class BasicPart(SeqRecord):
         else:
             raise ValueError("incorrect sequence used.")
 
-    def to_seqrec(self) -> SeqRecord:
-        """Create a SeqRecord instance. All relevant attributes are maintained."""
-        seqrec = SeqRecord(seq=self.seq, id=self.id)
-        for key in seqrec.__dict__.keys():
-            setattr(seqrec, key, self.__dict__[key])
-        return seqrec
-
-    def concentration(
+    def clip_mass(
         self,
-        stock: bool = True,
         clip_vol: float = 30,
         circular: bool = True,
         ndigit: int = None,
     ) -> Union[float, int]:
-        """Obtain the recommended concentration of part in ng/µL.
-
-        This can either be the recommended stock concentration of part or the final concentration of part in clip reactions.
+        """Obtain the recommended mass of part (ng) for a clip reaction.
 
         Args:
-            stock: If true, method returns the stock concentration required for adding 1 µL part to each clip reaction with volume = clip_vol. Else returns the suggested final concentration of part in a clip reaction (ng/µL).
             clip_vol: volume of the clip reaction (µL).
             circular: The part is circular or linear if False.
             ndigit: Refer to built-in round function documentation.
@@ -121,9 +122,57 @@ class BasicPart(SeqRecord):
             )
             / 1e6
         )
-        if stock:
-            return round(final_concentration * clip_vol, ndigit)
-        return round(final_concentration, ndigit)
+        return round(final_concentration * clip_vol, ndigit)
+
+    def domesticating_primers(
+        self,
+        seq_args: dict = 0,
+        global_args: dict = 0,
+        left_attrs: dict = None,
+        right_attrs: dict = None,
+        primer_pair: int = 0,
+    ) -> DomesticatingPrimers:
+        """Generate PCR primers and associated data for domesticating parts from various sources using PCR.
+
+        PCR primers contain iP and iS sequences, respectively, in addition to a region of homology for the intervening DNA sequence.
+        They can be used to domesticate parts from various sources, adding iP and iS sequences, enabling downstream BASIC assemblies.
+
+        Args:
+            seq_args: Refer to '_primer3py' method and primer3-py docs.
+            global_args: Refer to '_primer3py' method and primer3-py docs.
+            left_attrs: Additional attributes to assign to `left_primer` SeqRecord.
+            right_attrs: Additional attributes to assign to the `right_primer` SeqRecord.
+            primer_pair: Which of the returned primer3-py primer pairs to use. Defaults to 0 which is the primer pair with the lowest penalty value.
+
+        Returns:
+            DomesticatingPrimers: `left_primer` & `right_primer` attributes contain iP & iS sequences, respectively while `data` corresponds to primer regions homologous to template DNA.
+
+        Raises:
+            ValueError: If primer3 fails to identify suitable primer pairs.
+        """
+        p3py_out = self._primer3py(seq_args, global_args)
+        if p3py_out["PRIMER_PAIR_NUM_RETURNED"] == 0:
+            raise ValueError(
+                "primer3 returned no suitable primers. Refer to the primer3 documentation and change `seq_args` and/or `global_args`."
+            )
+        left_primer = p3_seqrec(p3py_out, "LEFT", left_attrs, primer_pair)
+        left_primer.seq = Seq(IP_STR) + left_primer.seq
+        right_primer = p3_seqrec(p3py_out, "RIGHT", right_attrs, primer_pair)
+        right_primer.seq = Seq(IS_STR).reverse_complement() + right_primer.seq
+        return DomesticatingPrimers(
+            left_primer,
+            right_primer,
+            data=dict(
+                item for item in p3py_out.items() if f"_{str(primer_pair)}_" in item[0]
+            ),
+        )
+
+    def to_seqrec(self) -> SeqRecord:
+        """Create a SeqRecord instance. All relevant attributes are maintained."""
+        seqrec = SeqRecord(seq=self.seq, id=self.id)
+        for key in seqrec.__dict__.keys():
+            setattr(seqrec, key, self.__dict__[key])
+        return seqrec
 
     def _find_iseq(
         self, seq: Seq, iseq_str: str, iseq_id: str = "integrated sequence"
@@ -171,6 +220,40 @@ class BasicPart(SeqRecord):
                 f"Sequence flanked by iP and iS sequences in {self.id} is {num_base_pairs} base pairs long, this is less than 90 base pairs which is incompatible with unligated linker removal during clip reaction purification."
             )
 
+    def _primer3py(
+        self,
+        seq_args: dict = None,
+        global_args: dict = None,
+    ) -> dict:
+        """Return dictionary of primer3-py results for sequence flanked by iP and iS sequences.
+
+        Notes:
+            Refer to the primer3-py documentation for further information.
+
+        Args:
+            seq_args: Primer3 sequence arguments. If left blank chooses suitable defaults.
+            global_args: Primer3 global arguments If left blank chooses suitable defaults.
+
+        """
+        template = str(self.basic_slice().seq)
+        input_seq_args = {
+            "SEQUENCE_TEMPLATE": template,
+        }
+        if seq_args:
+            input_seq_args.update(seq_args)
+        input_global_args = {
+            "PRIMER_TASK": "generic",
+            "PRIMER_PICK_LEFT_PRIMER": 1,
+            "PRIMER_PICK_RIGHT_PRIMER": 1,
+            "PRIMER_PRODUCT_SIZE_RANGE": [[len(template), len(template)]],
+        }
+        if global_args:
+            input_global_args.update(global_args)
+        return designPrimers(
+            input_seq_args,
+            input_global_args,
+        )
+
     @property
     def seq(self):
         return self._seq
@@ -194,39 +277,152 @@ class BasicPart(SeqRecord):
         return self.id == other.id and str(self.seq) == str(other.seq)
 
 
+@dataclass
+class LinkerHalfOligos:
+    """Oligos used by a linker half."""
+
+    long: Seq
+    adapter: Seq
+
+
+class _LinkerOligos:
+    """Description of oligonucleotides used to physically generate BasicLinker instances.
+
+    Not to be initiated directly but called by the BasicLinker class.
+
+    Attributes:
+        prefix: Oligos used to generate prefix linker half.
+        suffix: Oligos used to generate suffix linker half.
+
+    """
+
+    def __init__(self, seq: Seq, overhang_slice_params: Tuple[int, int]) -> None:
+        """Init function.
+
+        Args:
+            seq: Sequence of linker including scars.
+            overhang_slice_params: Parameters used to slice `seq` attribute returning overhang sequence.
+        """
+        self.prefix = LinkerHalfOligos(
+            long=seq[overhang_slice_params[0] :].reverse_complement(),
+            adapter=seq[
+                overhang_slice_params[1] : -1 * len(BasicLinker.DOWNSTREAM_SCAR)
+            ],
+        )
+        self.suffix = LinkerHalfOligos(
+            long=seq[2 : overhang_slice_params[1]],
+            adapter=seq[
+                len(BasicLinker.UPSTREAM_SCAR) : overhang_slice_params[0]
+            ].reverse_complement(),
+        )
+
+    def all_oligo_seqrecs(
+        self,
+        prefix_long_attrs: dict = None,
+        prefix_adapter_attrs: dict = None,
+        suffix_long_attrs: dict = None,
+        suffix_adapter_attrs: dict = None,
+    ) -> Tuple[SeqRecord, SeqRecord, SeqRecord, SeqRecord]:
+        """Return all oligos required to generate linker as SeqRecords.
+
+        Args:
+            prefix_long_attrs: Additional attributes to assign to `prefix.long` SeqRecord.
+            prefix_adapter_attrs: Additional attributes to assign to `prefix.adapter` SeqRecord.
+            suffix_long_attrs: Additional attributes to assign to `suffix.long` SeqRecord.
+            suffix_adapter_attrs: Additional attributes to assign to `suffix.adapter` SeqRecord.
+        """
+        return (
+            self._linker_oligo_seqrec(
+                seq=self.prefix.long,
+                id="prefix long",
+                additional_attrs=prefix_long_attrs,
+            ),
+            self._linker_oligo_seqrec(
+                seq=self.prefix.adapter,
+                id="prefix adapter",
+                additional_attrs=prefix_adapter_attrs,
+            ),
+            self._linker_oligo_seqrec(
+                seq=self.suffix.long,
+                id="suffix long",
+                additional_attrs=suffix_long_attrs,
+            ),
+            self._linker_oligo_seqrec(
+                seq=self.suffix.adapter,
+                id="suffix adapter",
+                additional_attrs=suffix_adapter_attrs,
+            ),
+        )
+
+    def _linker_oligo_seqrec(self, seq: Seq, id: str, additional_attrs: dict):
+        """Return a SeqRecord object for a linker oligonucleotide.
+
+        Args:
+            seq: Sequence of oligonucleotide.
+            id: ID attribute for oligonucleotide.
+            additional_attrs: Additional attributes to assign to oligonucleotide SeqRecord.
+
+        Notes:
+            By default assigns `OLIGO_ANNOTATIONS` to annoations attribute.
+        """
+        seqrec = SeqRecord(seq=seq, id=id, annotations=OLIGO_ANNOTATIONS)
+        if additional_attrs:
+            for key, value in additional_attrs.items():
+                setattr(seqrec, key, value)
+        return seqrec
+
+
 class BasicLinker(SeqRecord):
     """Class for BASIC DNA assembly linkers.
 
     A DNA sequence joined with other BasicLinkers via :py:class:`BasicPart`
-    instances when initialising :py:class:`BasicAssembly` objects.
+    instances when initializing :py:class:`BasicAssembly` objects.
 
     Attributes:
         prefix_id (str): The prefix half linker id.
         suffix_id (str): The suffix half linker id.
-        seq: Refer to Bio.SeqRecord.SeqRecord documentation.
+        seq : Refer to Bio.SeqRecord.SeqRecord documentation.
         id: Refer to Bio.SeqRecord.SeqRecord documentation.
-        features : Refer to Bio.SeqRecord.SeqRecord documentation.
+        name: This attribute is used to label the annotation in exported gb files or equivalent.
+        overhang_slice_params: Parameters used to slice `seq` attribute returning overhang sequence. Parameters are in the form (start, stop). Required to access `linker_oligos` attribute.
+        linker_oligos: Oligonucleotides used to physically generate linker.
 
     """
 
+    UPSTREAM_SCAR = "GGCTCG"
+    DOWNSTREAM_SCAR = "GTCC"
+
     @addargs2docs(CommonArgDocs.SEQREC_KWARGS)
     def __init__(
-        self, seq, id: str, prefix_id: str = None, suffix_id: str = None, **kwargs
+        self,
+        seq: Seq,
+        id: str,
+        name: str = None,
+        prefix_id: str = None,
+        suffix_id: str = None,
+        overhang_slice_params: Tuple[int, int] = None,
+        **kwargs,
     ):
         """Class for BASIC DNA assembly linkers.
 
         Args:
             seq : Refer to Bio.SeqRecord.SeqRecord documentation.
             id : Refer to Bio.SeqRecord.SeqRecord documentation.
+            name: This attribute is used to label the annotation in exported gb files or equivalent.
             prefix_id (optional): prefix id if known and not needing
-                generation, defaults to None.
+                generation, defaults to f"{self.name}-P".
             suffix_id (optional): suffix id if known and not needing
-                generation, defaults to None.
+                generation, defaults to f"{self.name}-S".
+            overhang_slice_params: Parameters used to slice `seq` attribute returning overhang sequence. Parameters are in the form (start, stop). Required to access `linker_oligos` attribute.
             kwargs:
         """
-        super().__init__(seq=seq, id=id, **kwargs)
+        self.seq = seq
+        super().__init__(seq=self.seq, id=id, name=name, **kwargs)
         self.prefix_id = self._assign_linker_half_id("prefix", prefix_id)
         self.suffix_id = self._assign_linker_half_id("suffix", suffix_id)
+        self.overhang_slice_params = overhang_slice_params
+        if overhang_slice_params:
+            self.linker_oligos = _LinkerOligos(self.seq, self.overhang_slice_params)
         self._linker_feature()
 
     def basic_slice(self):
@@ -243,7 +439,7 @@ class BasicLinker(SeqRecord):
             SeqFeature(
                 type="misc_feature",
                 location=FeatureLocation(2, len(self.seq), strand=+1),
-                qualifiers={"label": [str(self.id)]},
+                qualifiers={"label": [str(self.name)]},
             )
         )
 
@@ -265,6 +461,22 @@ class BasicLinker(SeqRecord):
             return f"{self.name}-S"
         return id
 
+    @property
+    def seq(self):
+        return self._seq
+
+    @seq.setter
+    def seq(self, value):
+        if value[: len(self.UPSTREAM_SCAR)] != self.UPSTREAM_SCAR:
+            raise ValueError(
+                f"BasicLinker class/subclass initiated with sequence lacking upstream scar: {self.UPSTREAM_SCAR}"
+            )
+        if value[-1 * len(self.DOWNSTREAM_SCAR) :] != self.DOWNSTREAM_SCAR:
+            raise ValueError(
+                f"BasicLinker class/subclass initiated with sequence lacking downstream scar: {self.DOWNSTREAM_SCAR}"
+            )
+        self._seq = value
+
     def __eq__(self, other: "BasicLinker") -> bool:
         if not isinstance(other, BasicLinker):
             raise TypeError(f"{other} is not a BasicLinker instance.")
@@ -274,8 +486,16 @@ class BasicLinker(SeqRecord):
 class BasicUTRRBSLinker(BasicLinker):
     """Sub-class of :py:class:`BasicLinker` for UTR-RBS linkers."""
 
-    def __init__(self, seq, id, prefix_id=None, suffix_id=None, **kwargs):
-        super().__init__(seq, id, prefix_id, suffix_id, **kwargs)
+    def __init__(
+        self,
+        seq: Seq,
+        id: str,
+        name: str,
+        prefix_id: str = None,
+        suffix_id: str = None,
+        **kwargs,
+    ):
+        super().__init__(seq, id, name, prefix_id, suffix_id, **kwargs)
         self.prefix_id = super()._assign_linker_half_id("prefix", prefix_id)
         self.suffix_id = f"UTR{self.name[3]}-S"
 
@@ -293,8 +513,6 @@ class BasicAssembly:
             object.
         parts_linkers: tuple of BasicPart and BasicLinker objects used to
             create this assembly.
-        clip_reactions: The :py:class:`ClipReaction` instances required
-            for BASIC assembly.
     """
 
     def __init__(self, id: str, *parts_linkers: Union[BasicPart, BasicLinker]):
@@ -317,7 +535,6 @@ class BasicAssembly:
             )
         self.id = id
         self.parts_linkers = parts_linkers
-        self.clip_reactions = self.return_clip_reactions()
 
     @addargs2docs(CommonArgDocs.SEQREC_KWARGS)
     def return_part(self, **kwargs) -> BasicPart:
@@ -379,7 +596,8 @@ class BasicAssembly:
         self._check_clip_reactions(clip_reactions)
         return tuple(clip_reactions)
 
-    def _check_clip_reactions(self, clip_reactions):
+    @staticmethod
+    def _check_clip_reactions(clip_reactions):
         """Checks `ClipReactions` are compatible e.g. same half linker not used
         multiple times.
 
@@ -440,6 +658,7 @@ class BasicAssembly:
                 f"BasicAssembly instance with id:{self.id}, initiated with an odd number of BasicPart/BasicLinker objects. An even number is a requirement."
             )
         self._parts_linkers = values
+        self._clip_reactions = self.return_clip_reactions()
 
 
 class ClipReaction:
@@ -503,7 +722,7 @@ class ClipReaction:
         Returns:
             string: hash of tuple containing, prefix_id, part_id and suffix_id
         """
-        return hash((self._prefix.id, self._part.id, self._suffix.id))
+        return hash((self._prefix.seq, self._part.seq, self._suffix.seq))
 
     def __eq__(self, other) -> bool:
         """The function test if an object `other` is equal to this `ClipReaction`.
