@@ -2,9 +2,12 @@
 
 Objects that peform various tasks but are not core objects of the
 basicsynbio package.
+Modules in this sub-package should not import from modules in other sub-packages of basicsynbio.
 """
 
 from .benchling import *
-from .biopython import _easy_seqrec
 from .biopython import *
+from .biopython import _easy_seqrec
 from .reportLab import *
+from .primer3py import *
+from .seva import *
